#include <iostream>
//#include <iomanip>
#include <qtoolbutton.h>
#include <qpainter.h>
#include <qlabel.h>
#include <qfiledialog.h>
#include <qmessagebox.h>
#include <qtooltip.h>
#include "imageviewer.h"

using namespace std;

int ImageDialog::numviewers = 0;

//==================================================================================
ImageViewer::ImageViewer( QWidget *parent, const char *name) : 
	QScrollView(parent, name, WStaticContents | WNoAutoErase)
{
	setResizePolicy( Manual );
	image = new QImage;
	pnmproc = new QProcess(this);
	connect( pnmproc, SIGNAL( wroteToStdin()  ), pnmproc, SLOT( closeStdin() ) );
	connect( pnmproc, SIGNAL( processExited() ), this, SLOT( pnmReady() ) );
}
//==================================================================================
ImageViewer::~ImageViewer()
{
	delete image;
}
//==================================================================================
//==================================================================================
void ImageViewer::setImage(const QByteArray &buffer)
{
	if( pnmproc->isRunning() ) {
		return;
	}
	pnmproc->clearArguments();
	pnmproc->addArgument("pfsoutppm");
	pnmproc->addArgument("-");
	if ( ! pnmproc->start() ) {
		cerr << "ImageViewer::setImage:Failure to start" << endl;
	}
	pnmproc->writeToStdin( buffer );
	
	return;
}
//==================================================================================
void ImageViewer::pnmReady()
{
	QByteArray pnmbuffer = pnmproc->readStdout();
	image->loadFromData(pnmbuffer);
	resizeContents(image->width(), image->height());
	repaintContents(0,0,image->width(), image->height(),false); 
	emit resizeRequest();
	return;
}
//==================================================================================
void ImageViewer::drawContents( QPainter * p, int clipx, int clipy, int clipw, int cliph )
{
	if( ! image->isNull() )
		p->drawImage( clipx, clipy, *image, clipx, clipy, clipw, cliph );
	return;
}
//==================================================================================
QSize ImageViewer::sizeHint() const
{
	if( ! image->isNull() )
		return QSize(image->width() + 10, image->height() + 35);
	return QWidget::sizeHint();
}
//==================================================================================
bool ImageViewer::save(const QString fname) 
{
	QString fmt = QImage::imageFormat(fname);
	cout << "ImageViewer::save:  " << fname << ", fmt:  '" << fmt << "'" << flush;
	fmt = "PNG";
	bool rc = image->save(fname,fmt);
	cout << ", rc: " << rc << endl;
	return rc; 
}
//==================================================================================
//**********************************************************************************
//==================================================================================
ImageDialog::ImageDialog( QWidget *parent, const char *name) : 
	QMainWindow(parent, name)
{
	imageviewer = new ImageViewer(this);
	setCentralWidget( imageviewer );
	connect( imageviewer, SIGNAL( resizeRequest()), this, SLOT( resizing() ) );
	
	
	QToolBar *toolBar = new QToolBar( this );
	
	//savebutton = new QPushButton(QPixmap::fromMimeSource("images/filesave.png"), "", toolBar );
	savebutton = new QPushButton(QPixmap::fromMimeSource("filesave.png"), "", toolBar );
	connect( savebutton, SIGNAL( clicked()), this, SLOT( saveImage()));
	QToolTip::add(savebutton, "Save current image to a file");
	
	QLabel *gammatag = new QLabel("Gamma", toolBar);
	gammaslider = new QSlider(0,100,10, 0, Qt::Horizontal, toolBar);
	gammaslider->setMinimumWidth(150);
	
	gammalineedit = new QLineEdit("1.00", toolBar);
	gammalineedit->setAlignment(Qt::AlignRight);
	gammalineedit->setReadOnly(true);
	gammalineedit->setMaximumWidth(50);
	gammagang = new Gang(gammaslider, gammalineedit, 0, 3, 1);
	curr_gamma = 1;
	QToolTip::add( gammaslider, "Adjust gamma for this image");
	QToolTip::add( gammatag, "Adjust gamma for this image");
	
	connect( gammaslider, SIGNAL( sliderMoved(int)), this, SLOT( setGamma(int)));
	
	//QPushButton *resetbutton = new QPushButton(QPixmap::fromMimeSource("images/undo.png"),"",toolBar);
	QPushButton *resetbutton = new QPushButton(QPixmap::fromMimeSource("undo.png"),"",toolBar);
	connect( resetbutton, SIGNAL( clicked() ), this, SLOT( resetGamma() ));
	QToolTip::add(resetbutton, "Reset to gamma = 1");
	
	//QPushButton *closebutton = new QPushButton(QPixmap::fromMimeSource("images/close.png"), "",toolBar);
	QPushButton *closebutton = new QPushButton(QPixmap::fromMimeSource("close.png"), "",toolBar);
	connect( closebutton, SIGNAL( clicked() ), this, SLOT( close() ));
	QToolTip::add(closebutton, "Close this window without quiting qpfstmo");
	
	workinglabel = new QLabel( toolBar);
	
	pfsproc = new QProcess(this);
	connect( pfsproc, SIGNAL( wroteToStdin()  ), pfsproc, SLOT( closeStdin() ) );
	connect( pfsproc, SIGNAL( processExited() ), this, SLOT( pfsReady() ) );
	
	numviewers++;
}
//==================================================================================
ImageDialog::~ImageDialog()
{
	cout << "ImageDialog::~ImageDialog(): numviewers:  " << numviewers << endl;
}
//==================================================================================
void ImageDialog::set(const QByteArray &pfsin, QString cmd, QString args, QString caption, QString fname)
{
	setCaption(caption);
	filename = fname;
	gammagang->setDefault();
	if( cmd == "" ) {
		// No tone mapping for this buffer, just display
		tmbuffer = pfsin;
		tmbuffer.detach(); // Make this a local copy
		imageviewer->setImage(tmbuffer);
	}
	else {
		// Start with doing requested tone mapping
		tmbuffer.resize(0);
		savebutton->setEnabled(false);
		gammaslider->setEnabled(false);
		workinglabel->setEnabled(true);
		workinglabel->setText("<b>Working</b>");
		startProc(pfsin, cmd, args);
	}
	return;
}

//==================================================================================
void ImageDialog::pfsReady()
{
	if( tmbuffer.size() == 0) {
		// process call was for initial tone mapping, save the result
		tmbuffer = pfsproc->readStdout();
		savebutton->setEnabled(true);
		gammaslider->setEnabled(true);
		workinglabel->setEnabled(false);
		workinglabel->setText("");
		imageviewer->setImage( tmbuffer );
	}
	else {
		// was a gamma adjust, no need to save the result
		QByteArray gammabuffer = pfsproc->readStdout();
		imageviewer->setImage( gammabuffer );
		// If we are not up to date, start new gamma adjust
		if( curr_gamma != gammagang->v() )
			setGamma(1);
	}
	return;
}
//==================================================================================
void ImageDialog::resizing()
{
	resize(imageviewer->sizeHint());
}
//==================================================================================
void ImageDialog::setGamma(int value)
{
	curr_gamma = gammagang->v();
	startProc(tmbuffer,"pfsgamma", QString("-g %1").arg(curr_gamma));
	return;
}
//==================================================================================
void ImageDialog::startProc(const QByteArray & pfsbuffer, QString cmd, QString args)
{
	if( pfsproc->isRunning() ) {
		return;
	}
	else {
		pfsproc->setArguments( QStringList::split(QString(" "), cmd + " " + args) );
		if ( ! pfsproc->start() ) {
			cerr << "ImageDialog::startProc:Failure to start" << endl;
		}
		pfsproc->writeToStdin( pfsbuffer );
	}
	return;
}
//==================================================================================
void ImageDialog::resetGamma()
{
	gammagang->setDefault();
	if( curr_gamma != gammagang->v() )
		setGamma(1);
	return;
}
//==================================================================================
/* For reasons unknown to me, I can only save as PNG. -ngp
*/
void ImageDialog::saveImage()
{	
	QString cwd = ".";
	QString filetypes;
	filetypes += "PNG (*.png);;";
	filetypes += "JPEG (*.jpeg,*.jpg);;";
	filetypes += "TIFF (*.tiff,*.tif);;";
	filetypes += "PNM, PPM  (*.ppm,*.pnm);;";
	filetypes += "GIF (*.gif);;";
	filetypes += "BMP (*.bmp);;";
	filetypes += "All Files (*)";
	QString outfname = QFileDialog::getSaveFileName(filename, 
					filetypes,this,
					"Choose", "Save Image as");
	if( ! outfname.isEmpty() ) {
		if( ! imageviewer->save(outfname) ) {
			QMessageBox::warning(this,"","Failed to save to <b>" + outfname + "</b>",
				QMessageBox::Ok, QMessageBox::NoButton);
		}
	}

}
